import { SvelteComponent, init, safe_not_equal, create_slot, element, attr, insert, update_slot, transition_in, transition_out, detach, binding_callbacks, create_component, mount_component, destroy_component, empty, space, set_style, append, listen, group_outros, check_outros, update_keyed_each, outro_and_destroy_block, compute_slots } from 'svelte/internal';
import { createEventDispatcher, onMount, afterUpdate, onDestroy } from 'svelte';

/**
 * virtual list core calculating center
 */

const DIRECTION_TYPE = {
    FRONT: "FRONT", // scroll up or left
    BEHIND: "BEHIND", // scroll down or right
};
const CALC_TYPE = {
    INIT: "INIT",
    FIXED: "FIXED",
    DYNAMIC: "DYNAMIC",
};
const LEADING_BUFFER = 2;

class Virtual {
    param
    callUpdate
    firstRangeTotalSize = 0
    firstRangeAverageSize = 0
    lastCalcIndex = 0
    fixedSizeValue = 0
    calcType = CALC_TYPE.INIT
    offset = 0
    direction = ""
    range

    constructor(param, callUpdate) {
        this.init(param, callUpdate);
    }

    init(param, callUpdate) {
        // param data
        this.param = param;
        this.callUpdate = callUpdate;

        // size data
        this.sizes = new Map();
        this.firstRangeTotalSize = 0;
        this.firstRangeAverageSize = 0;
        this.lastCalcIndex = 0;
        this.fixedSizeValue = 0;
        this.calcType = CALC_TYPE.INIT;

        // scroll data
        this.offset = 0;
        this.direction = "";

        // range data
        this.range = Object.create(null);
        if (param) {
            this.checkRange(0, param.keeps - 1);
        }

        // benchmark example data
        // this.__bsearchCalls = 0
        // this.__getIndexOffsetCalls = 0
    }

    destroy() {
        this.init(null, null);
    }

    // return current render range
    getRange() {
        const range = Object.create(null);
        range.start = this.range.start;
        range.end = this.range.end;
        range.padFront = this.range.padFront;
        range.padBehind = this.range.padBehind;
        return range
    }

    isBehind() {
        return this.direction === DIRECTION_TYPE.BEHIND
    }

    isFront() {
        return this.direction === DIRECTION_TYPE.FRONT
    }

    // return start index offset
    getOffset(start) {
        return (start < 1 ? 0 : this.getIndexOffset(start)) + this.param.slotHeaderSize
    }

    updateParam(key, value) {
        if (this.param && (key in this.param)) {
            // if uniqueIds change, find out deleted id and remove from size map
            if (key === "uniqueIds") {
                this.sizes.forEach((v, key) => {
                    if (!value.includes(key)) {
                        this.sizes.delete(key);
                    }
                });
            }
            this.param[key] = value;
        }
    }

    // save each size map by id
    saveSize(id, size) {
        this.sizes.set(id, size);

        // we assume size type is fixed at the beginning and remember first size value
        // if there is no size value different from this at next coming saving
        // we think it's a fixed size list, otherwise is dynamic size list
        if (this.calcType === CALC_TYPE.INIT) {
            this.fixedSizeValue = size;
            this.calcType = CALC_TYPE.FIXED;
        } else if (this.calcType === CALC_TYPE.FIXED && this.fixedSizeValue !== size) {
            this.calcType = CALC_TYPE.DYNAMIC;
            // it's no use at all
            delete this.fixedSizeValue;
        }

        // calculate the average size only in the first range
        if (this.calcType !== CALC_TYPE.FIXED && typeof this.firstRangeTotalSize !== "undefined") {
            if (this.sizes.size < Math.min(this.param.keeps, this.param.uniqueIds.length)) {
                this.firstRangeTotalSize = [...this.sizes.values()].reduce((acc, val) => acc + val, 0);
                this.firstRangeAverageSize = Math.round(this.firstRangeTotalSize / this.sizes.size);
            } else {
                // it's done using
                delete this.firstRangeTotalSize;
            }
        }
    }

    // in some special situation (e.g. length change) we need to update in a row
    // try going to render next range by a leading buffer according to current direction
    handleDataSourcesChange() {
        let start = this.range.start;

        if (this.isFront()) {
            start = start - LEADING_BUFFER;
        } else if (this.isBehind()) {
            start = start + LEADING_BUFFER;
        }

        start = Math.max(start, 0);

        this.updateRange(this.range.start, this.getEndByStart(start));
    }

    // when slot size change, we also need force update
    handleSlotSizeChange() {
        this.handleDataSourcesChange();
    }

    // calculating range on scroll
    handleScroll(offset) {
        this.direction = offset < this.offset ? DIRECTION_TYPE.FRONT : DIRECTION_TYPE.BEHIND;
        this.offset = offset;

        if (!this.param) {
            return
        }

        if (this.direction === DIRECTION_TYPE.FRONT) {
            this.handleFront();
        } else if (this.direction === DIRECTION_TYPE.BEHIND) {
            this.handleBehind();
        }
    }

    // ----------- public method end -----------

    handleFront() {
        const overs = this.getScrollOvers();
        // should not change range if start doesn't exceed overs
        if (overs > this.range.start) {
            return
        }

        // move up start by a buffer length, and make sure its safety
        const start = Math.max(overs - this.param.buffer, 0);
        this.checkRange(start, this.getEndByStart(start));
    }

    handleBehind() {
        const overs = this.getScrollOvers();
        // range should not change if scroll overs within buffer
        if (overs < this.range.start + this.param.buffer) {
            return
        }

        this.checkRange(overs, this.getEndByStart(overs));
    }

    // return the pass overs according to current scroll offset
    getScrollOvers() {
        // if slot header exist, we need subtract its size
        const offset = this.offset - this.param.slotHeaderSize;
        if (offset <= 0) {
            return 0
        }

        // if is fixed type, that can be easily
        if (this.isFixedType()) {
            return Math.floor(offset / this.fixedSizeValue)
        }

        let low = 0;
        let middle = 0;
        let middleOffset = 0;
        let high = this.param.uniqueIds.length;

        while (low <= high) {
            // this.__bsearchCalls++
            middle = low + Math.floor((high - low) / 2);
            middleOffset = this.getIndexOffset(middle);

            if (middleOffset === offset) {
                return middle
            } else if (middleOffset < offset) {
                low = middle + 1;
            } else if (middleOffset > offset) {
                high = middle - 1;
            }
        }

        return low > 0 ? --low : 0
    }

    // return a scroll offset from given index, can efficiency be improved more here?
    // although the call frequency is very high, its only a superposition of numbers
    getIndexOffset(givenIndex) {
        if (!givenIndex) {
            return 0
        }

        let offset = 0;
        let indexSize = 0;
        for (let index = 0; index < givenIndex; index++) {
            // this.__getIndexOffsetCalls++
            indexSize = this.sizes.get(this.param.uniqueIds[index]);
            offset = offset + (typeof indexSize === "number" ? indexSize : this.getEstimateSize());
        }

        // remember last calculate index
        this.lastCalcIndex = Math.max(this.lastCalcIndex, givenIndex - 1);
        this.lastCalcIndex = Math.min(this.lastCalcIndex, this.getLastIndex());

        return offset
    }

    // is fixed size type
    isFixedType() {
        return this.calcType === CALC_TYPE.FIXED
    }

    // return the real last index
    getLastIndex() {
        return this.param.uniqueIds.length - 1
    }

    // in some conditions range is broke, we need correct it
    // and then decide whether need update to next range
    checkRange(start, end) {
        const keeps = this.param.keeps;
        const total = this.param.uniqueIds.length;

        // data less than keeps, render all
        if (total <= keeps) {
            start = 0;
            end = this.getLastIndex();
        } else if (end - start < keeps - 1) {
            // if range length is less than keeps, correct it base on end
            start = end - keeps + 1;
        }

        if (this.range.start !== start) {
            this.updateRange(start, end);
        }
    }

    // setting to a new range and rerender
    updateRange(start, end) {
        this.range.start = start;
        this.range.end = end;
        this.range.padFront = this.getPadFront();
        this.range.padBehind = this.getPadBehind();
        this.callUpdate(this.getRange());
    }

    // return end base on start
    getEndByStart(start) {
        const theoryEnd = start + this.param.keeps - 1;
        const truelyEnd = Math.min(theoryEnd, this.getLastIndex());
        return truelyEnd
    }

    // return total front offset
    getPadFront() {
        if (this.isFixedType()) {
            return this.fixedSizeValue * this.range.start
        } else {
            return this.getIndexOffset(this.range.start)
        }
    }

    // return total behind offset
    getPadBehind() {
        const end = this.range.end;
        const lastIndex = this.getLastIndex();

        if (this.isFixedType()) {
            return (lastIndex - end) * this.fixedSizeValue
        }

        // if it's all calculated, return the exactly offset
        if (this.lastCalcIndex === lastIndex) {
            return this.getIndexOffset(lastIndex) - this.getIndexOffset(end)
        } else {
            // if not, use a estimated value
            return (lastIndex - end) * this.getEstimateSize()
        }
    }

    // get the item estimate size
    getEstimateSize() {
        return this.isFixedType() ? this.fixedSizeValue : (this.firstRangeAverageSize || this.param.estimateSize)
    }
}

/* src/Item.svelte generated by Svelte v3.38.2 */

function create_fragment$1(ctx) {
	let div;
	let current;
	const default_slot_template = /*#slots*/ ctx[5].default;
	const default_slot = create_slot(default_slot_template, ctx, /*$$scope*/ ctx[4], null);

	return {
		c() {
			div = element("div");
			if (default_slot) default_slot.c();
			attr(div, "class", "virtual-scroll-item");
		},
		m(target, anchor) {
			insert(target, div, anchor);

			if (default_slot) {
				default_slot.m(div, null);
			}

			/*div_binding*/ ctx[6](div);
			current = true;
		},
		p(ctx, [dirty]) {
			if (default_slot) {
				if (default_slot.p && (!current || dirty & /*$$scope*/ 16)) {
					update_slot(default_slot, default_slot_template, ctx, /*$$scope*/ ctx[4], dirty, null, null);
				}
			}
		},
		i(local) {
			if (current) return;
			transition_in(default_slot, local);
			current = true;
		},
		o(local) {
			transition_out(default_slot, local);
			current = false;
		},
		d(detaching) {
			if (detaching) detach(div);
			if (default_slot) default_slot.d(detaching);
			/*div_binding*/ ctx[6](null);
		}
	};
}

function instance$1($$self, $$props, $$invalidate) {
	let { $$slots: slots = {}, $$scope } = $$props;
	let { horizontal = false } = $$props;
	let { uniqueKey } = $$props;
	let { type = "item" } = $$props;
	let resizeObserver;
	let itemDiv;
	let previousSize;
	const dispatch = createEventDispatcher();
	const shapeKey = horizontal ? "offsetWidth" : "offsetHeight";

	onMount(() => {
		if (typeof ResizeObserver !== "undefined") {
			resizeObserver = new ResizeObserver(dispatchSizeChange);
			resizeObserver.observe(itemDiv);
		}
	});

	afterUpdate(dispatchSizeChange);

	onDestroy(() => {
		if (resizeObserver) {
			resizeObserver.disconnect();
			resizeObserver = null;
		}
	});

	function dispatchSizeChange() {
		const size = itemDiv ? itemDiv[shapeKey] : 0;
		if (size === previousSize) return;
		previousSize = size;
		dispatch("resize", { id: uniqueKey, size, type });
	}

	function div_binding($$value) {
		binding_callbacks[$$value ? "unshift" : "push"](() => {
			itemDiv = $$value;
			$$invalidate(0, itemDiv);
		});
	}

	$$self.$$set = $$props => {
		if ("horizontal" in $$props) $$invalidate(1, horizontal = $$props.horizontal);
		if ("uniqueKey" in $$props) $$invalidate(2, uniqueKey = $$props.uniqueKey);
		if ("type" in $$props) $$invalidate(3, type = $$props.type);
		if ("$$scope" in $$props) $$invalidate(4, $$scope = $$props.$$scope);
	};

	return [itemDiv, horizontal, uniqueKey, type, $$scope, slots, div_binding];
}

class Item extends SvelteComponent {
	constructor(options) {
		super();
		init(this, options, instance$1, create_fragment$1, safe_not_equal, { horizontal: 1, uniqueKey: 2, type: 3 });
	}
}

/* src/VirtualScroll.svelte generated by Svelte v3.38.2 */
const get_footer_slot_changes = dirty => ({ data: dirty[0] & /*displayItems*/ 4 });
const get_footer_slot_context = ctx => ({ data: /*dataItem*/ ctx[39] });

function get_each_context(ctx, list, i) {
	const child_ctx = ctx.slice();
	child_ctx[39] = list[i];
	return child_ctx;
}

const get_default_slot_changes = dirty => ({ data: dirty[0] & /*displayItems*/ 4 });
const get_default_slot_context = ctx => ({ data: /*dataItem*/ ctx[39] });
const get_header_slot_changes = dirty => ({ data: dirty[0] & /*displayItems*/ 4 });
const get_header_slot_context = ctx => ({ data: /*dataItem*/ ctx[39] });

// (265:4) {#if $$slots.header}
function create_if_block_1(ctx) {
	let item;
	let current;

	item = new Item({
			props: {
				type: "slot",
				uniqueKey: "header",
				$$slots: { default: [create_default_slot_2] },
				$$scope: { ctx }
			}
		});

	item.$on("resize", /*onItemResized*/ ctx[6]);

	return {
		c() {
			create_component(item.$$.fragment);
		},
		m(target, anchor) {
			mount_component(item, target, anchor);
			current = true;
		},
		p(ctx, dirty) {
			const item_changes = {};

			if (dirty[0] & /*$$scope, displayItems*/ 536870916) {
				item_changes.$$scope = { dirty, ctx };
			}

			item.$set(item_changes);
		},
		i(local) {
			if (current) return;
			transition_in(item.$$.fragment, local);
			current = true;
		},
		o(local) {
			transition_out(item.$$.fragment, local);
			current = false;
		},
		d(detaching) {
			destroy_component(item, detaching);
		}
	};
}

// (266:8) <Item on:resize={onItemResized} type="slot" uniqueKey="header">
function create_default_slot_2(ctx) {
	let current;
	const header_slot_template = /*#slots*/ ctx[26].header;
	const header_slot = create_slot(header_slot_template, ctx, /*$$scope*/ ctx[29], get_header_slot_context);

	return {
		c() {
			if (header_slot) header_slot.c();
		},
		m(target, anchor) {
			if (header_slot) {
				header_slot.m(target, anchor);
			}

			current = true;
		},
		p(ctx, dirty) {
			if (header_slot) {
				if (header_slot.p && (!current || dirty[0] & /*$$scope, displayItems*/ 536870916)) {
					update_slot(header_slot, header_slot_template, ctx, /*$$scope*/ ctx[29], dirty, get_header_slot_changes, get_header_slot_context);
				}
			}
		},
		i(local) {
			if (current) return;
			transition_in(header_slot, local);
			current = true;
		},
		o(local) {
			transition_out(header_slot, local);
			current = false;
		},
		d(detaching) {
			if (header_slot) header_slot.d(detaching);
		}
	};
}

// (272:12) <Item                     on:resize={onItemResized}                     uniqueKey={dataItem[key]}                     horizontal={isHorizontal}                     type="item">
function create_default_slot_1(ctx) {
	let t;
	let current;
	const default_slot_template = /*#slots*/ ctx[26].default;
	const default_slot = create_slot(default_slot_template, ctx, /*$$scope*/ ctx[29], get_default_slot_context);

	return {
		c() {
			if (default_slot) default_slot.c();
			t = space();
		},
		m(target, anchor) {
			if (default_slot) {
				default_slot.m(target, anchor);
			}

			insert(target, t, anchor);
			current = true;
		},
		p(ctx, dirty) {
			if (default_slot) {
				if (default_slot.p && (!current || dirty[0] & /*$$scope, displayItems*/ 536870916)) {
					update_slot(default_slot, default_slot_template, ctx, /*$$scope*/ ctx[29], dirty, get_default_slot_changes, get_default_slot_context);
				}
			}
		},
		i(local) {
			if (current) return;
			transition_in(default_slot, local);
			current = true;
		},
		o(local) {
			transition_out(default_slot, local);
			current = false;
		},
		d(detaching) {
			if (default_slot) default_slot.d(detaching);
			if (detaching) detach(t);
		}
	};
}

// (271:8) {#each displayItems as dataItem (dataItem[key])}
function create_each_block(key_2, ctx) {
	let first;
	let item;
	let current;

	item = new Item({
			props: {
				uniqueKey: /*dataItem*/ ctx[39][/*key*/ ctx[0]],
				horizontal: /*isHorizontal*/ ctx[1],
				type: "item",
				$$slots: { default: [create_default_slot_1] },
				$$scope: { ctx }
			}
		});

	item.$on("resize", /*onItemResized*/ ctx[6]);

	return {
		key: key_2,
		first: null,
		c() {
			first = empty();
			create_component(item.$$.fragment);
			this.first = first;
		},
		m(target, anchor) {
			insert(target, first, anchor);
			mount_component(item, target, anchor);
			current = true;
		},
		p(new_ctx, dirty) {
			ctx = new_ctx;
			const item_changes = {};
			if (dirty[0] & /*displayItems, key*/ 5) item_changes.uniqueKey = /*dataItem*/ ctx[39][/*key*/ ctx[0]];
			if (dirty[0] & /*isHorizontal*/ 2) item_changes.horizontal = /*isHorizontal*/ ctx[1];

			if (dirty[0] & /*$$scope, displayItems*/ 536870916) {
				item_changes.$$scope = { dirty, ctx };
			}

			item.$set(item_changes);
		},
		i(local) {
			if (current) return;
			transition_in(item.$$.fragment, local);
			current = true;
		},
		o(local) {
			transition_out(item.$$.fragment, local);
			current = false;
		},
		d(detaching) {
			if (detaching) detach(first);
			destroy_component(item, detaching);
		}
	};
}

// (281:4) {#if $$slots.footer}
function create_if_block(ctx) {
	let item;
	let current;

	item = new Item({
			props: {
				type: "slot",
				uniqueKey: "footer",
				$$slots: { default: [create_default_slot] },
				$$scope: { ctx }
			}
		});

	item.$on("resize", /*onItemResized*/ ctx[6]);

	return {
		c() {
			create_component(item.$$.fragment);
		},
		m(target, anchor) {
			mount_component(item, target, anchor);
			current = true;
		},
		p(ctx, dirty) {
			const item_changes = {};

			if (dirty[0] & /*$$scope, displayItems*/ 536870916) {
				item_changes.$$scope = { dirty, ctx };
			}

			item.$set(item_changes);
		},
		i(local) {
			if (current) return;
			transition_in(item.$$.fragment, local);
			current = true;
		},
		o(local) {
			transition_out(item.$$.fragment, local);
			current = false;
		},
		d(detaching) {
			destroy_component(item, detaching);
		}
	};
}

// (282:8) <Item on:resize={onItemResized} type="slot" uniqueKey="footer">
function create_default_slot(ctx) {
	let current;
	const footer_slot_template = /*#slots*/ ctx[26].footer;
	const footer_slot = create_slot(footer_slot_template, ctx, /*$$scope*/ ctx[29], get_footer_slot_context);

	return {
		c() {
			if (footer_slot) footer_slot.c();
		},
		m(target, anchor) {
			if (footer_slot) {
				footer_slot.m(target, anchor);
			}

			current = true;
		},
		p(ctx, dirty) {
			if (footer_slot) {
				if (footer_slot.p && (!current || dirty[0] & /*$$scope, displayItems*/ 536870916)) {
					update_slot(footer_slot, footer_slot_template, ctx, /*$$scope*/ ctx[29], dirty, get_footer_slot_changes, get_footer_slot_context);
				}
			}
		},
		i(local) {
			if (current) return;
			transition_in(footer_slot, local);
			current = true;
		},
		o(local) {
			transition_out(footer_slot, local);
			current = false;
		},
		d(detaching) {
			if (footer_slot) footer_slot.d(detaching);
		}
	};
}

function create_fragment(ctx) {
	let div2;
	let t0;
	let div0;
	let each_blocks = [];
	let each_1_lookup = new Map();
	let t1;
	let t2;
	let div1;
	let current;
	let mounted;
	let dispose;
	let if_block0 = /*$$slots*/ ctx[8].header && create_if_block_1(ctx);
	let each_value = /*displayItems*/ ctx[2];
	const get_key = ctx => /*dataItem*/ ctx[39][/*key*/ ctx[0]];

	for (let i = 0; i < each_value.length; i += 1) {
		let child_ctx = get_each_context(ctx, each_value, i);
		let key = get_key(child_ctx);
		each_1_lookup.set(key, each_blocks[i] = create_each_block(key, child_ctx));
	}

	let if_block1 = /*$$slots*/ ctx[8].footer && create_if_block(ctx);

	return {
		c() {
			div2 = element("div");
			if (if_block0) if_block0.c();
			t0 = space();
			div0 = element("div");

			for (let i = 0; i < each_blocks.length; i += 1) {
				each_blocks[i].c();
			}

			t1 = space();
			if (if_block1) if_block1.c();
			t2 = space();
			div1 = element("div");
			set_style(div0, "padding", /*paddingStyle*/ ctx[3]);
			attr(div1, "class", "shepherd");
			set_style(div1, "width", /*isHorizontal*/ ctx[1] ? "0px" : "100%");
			set_style(div1, "height", /*isHorizontal*/ ctx[1] ? "100%" : "0px");
			set_style(div2, "overflow-y", "auto");
			set_style(div2, "height", "inherit");
		},
		m(target, anchor) {
			insert(target, div2, anchor);
			if (if_block0) if_block0.m(div2, null);
			append(div2, t0);
			append(div2, div0);

			for (let i = 0; i < each_blocks.length; i += 1) {
				each_blocks[i].m(div0, null);
			}

			append(div2, t1);
			if (if_block1) if_block1.m(div2, null);
			append(div2, t2);
			append(div2, div1);
			/*div1_binding*/ ctx[27](div1);
			/*div2_binding*/ ctx[28](div2);
			current = true;

			if (!mounted) {
				dispose = listen(div2, "scroll", /*onScroll*/ ctx[7]);
				mounted = true;
			}
		},
		p(ctx, dirty) {
			if (/*$$slots*/ ctx[8].header) {
				if (if_block0) {
					if_block0.p(ctx, dirty);

					if (dirty[0] & /*$$slots*/ 256) {
						transition_in(if_block0, 1);
					}
				} else {
					if_block0 = create_if_block_1(ctx);
					if_block0.c();
					transition_in(if_block0, 1);
					if_block0.m(div2, t0);
				}
			} else if (if_block0) {
				group_outros();

				transition_out(if_block0, 1, 1, () => {
					if_block0 = null;
				});

				check_outros();
			}

			if (dirty[0] & /*displayItems, key, isHorizontal, onItemResized, $$scope*/ 536870983) {
				each_value = /*displayItems*/ ctx[2];
				group_outros();
				each_blocks = update_keyed_each(each_blocks, dirty, get_key, 1, ctx, each_value, each_1_lookup, div0, outro_and_destroy_block, create_each_block, null, get_each_context);
				check_outros();
			}

			if (!current || dirty[0] & /*paddingStyle*/ 8) {
				set_style(div0, "padding", /*paddingStyle*/ ctx[3]);
			}

			if (/*$$slots*/ ctx[8].footer) {
				if (if_block1) {
					if_block1.p(ctx, dirty);

					if (dirty[0] & /*$$slots*/ 256) {
						transition_in(if_block1, 1);
					}
				} else {
					if_block1 = create_if_block(ctx);
					if_block1.c();
					transition_in(if_block1, 1);
					if_block1.m(div2, t2);
				}
			} else if (if_block1) {
				group_outros();

				transition_out(if_block1, 1, 1, () => {
					if_block1 = null;
				});

				check_outros();
			}

			if (!current || dirty[0] & /*isHorizontal*/ 2) {
				set_style(div1, "width", /*isHorizontal*/ ctx[1] ? "0px" : "100%");
			}

			if (!current || dirty[0] & /*isHorizontal*/ 2) {
				set_style(div1, "height", /*isHorizontal*/ ctx[1] ? "100%" : "0px");
			}
		},
		i(local) {
			if (current) return;
			transition_in(if_block0);

			for (let i = 0; i < each_value.length; i += 1) {
				transition_in(each_blocks[i]);
			}

			transition_in(if_block1);
			current = true;
		},
		o(local) {
			transition_out(if_block0);

			for (let i = 0; i < each_blocks.length; i += 1) {
				transition_out(each_blocks[i]);
			}

			transition_out(if_block1);
			current = false;
		},
		d(detaching) {
			if (detaching) detach(div2);
			if (if_block0) if_block0.d();

			for (let i = 0; i < each_blocks.length; i += 1) {
				each_blocks[i].d();
			}

			if (if_block1) if_block1.d();
			/*div1_binding*/ ctx[27](null);
			/*div2_binding*/ ctx[28](null);
			mounted = false;
			dispose();
		}
	};
}

function instance($$self, $$props, $$invalidate) {
	let { $$slots: slots = {}, $$scope } = $$props;
	const $$slots = compute_slots(slots);
	let { key = "id" } = $$props;
	let { data } = $$props;
	let { keeps = 30 } = $$props;
	let { estimateSize = 50 } = $$props;
	let { isHorizontal = false } = $$props;
	let { start = 0 } = $$props;
	let { offset = 0 } = $$props;
	let { pageMode = false } = $$props;
	let { topThreshold = 0 } = $$props;
	let { bottomThreshold = 0 } = $$props;
	let displayItems = [];
	let paddingStyle;
	let directionKey = isHorizontal ? "scrollLeft" : "scrollTop";
	let range = null;

	let virtual = new Virtual({
			slotHeaderSize: 0,
			slotFooterSize: 0,
			keeps,
			estimateSize,
			buffer: Math.round(keeps / 3), // recommend for a third of keeps
			uniqueIds: getUniqueIdFromDataSources()
		},
	onRangeChanged);

	let root;
	let shepherd;
	const dispatch = createEventDispatcher();

	function getSize(id) {
		return virtual.sizes.get(id);
	}

	function getSizes() {
		return virtual.sizes.size;
	}

	function getOffset() {
		if (pageMode) {
			return document.documentElement[directionKey] || document.body[directionKey];
		} else {
			return root ? Math.ceil(root[directionKey]) : 0;
		}
	}

	function getClientSize() {
		const key = isHorizontal ? "clientWidth" : "clientHeight";

		if (pageMode) {
			return document.documentElement[key] || document.body[key];
		} else {
			return root ? Math.ceil(root[key]) : 0;
		}
	}

	function getScrollSize() {
		const key = isHorizontal ? "scrollWidth" : "scrollHeight";

		if (pageMode) {
			return document.documentElement[key] || document.body[key];
		} else {
			return root ? Math.ceil(root[key]) : 0;
		}
	}

	function updatePageModeFront() {
		if (root) {
			const rect = root.getBoundingClientRect();
			const { defaultView } = root.ownerDocument;

			const offsetFront = isHorizontal
			? rect.left + defaultView.pageXOffset
			: rect.top + defaultView.pageYOffset;

			virtual.updateParam("slotHeaderSize", offsetFront);
		}
	}

	function scrollToOffset(offset) {
		if (pageMode) {
			document.body[directionKey] = offset;
			document.documentElement[directionKey] = offset;
		} else if (root) {
			$$invalidate(4, root[directionKey] = offset, root);
		}
	}

	function scrollToIndex(index) {
		if (index >= data.length - 1) {
			scrollToBottom();
		} else {
			const offset = virtual.getOffset(index);
			scrollToOffset(offset);
		}
	}

	function scrollToBottom() {
		if (shepherd) {
			const offset = shepherd[isHorizontal ? "offsetLeft" : "offsetTop"];
			scrollToOffset(offset);

			// check if it's really scrolled to the bottom
			// maybe list doesn't render and calculate to last range
			// so we need retry in next event loop until it really at bottom
			setTimeout(
				() => {
					if (getOffset() + getClientSize() + 1 < getScrollSize()) {
						scrollToBottom();
					}
				},
				3
			);
		}
	}

	onMount(() => {
		if (start) {
			scrollToIndex(start);
		} else if (offset) {
			scrollToOffset(offset);
		}

		if (pageMode) {
			updatePageModeFront();
			document.addEventListener("scroll", onScroll, { passive: false });
		}
	});

	onDestroy(() => {
		virtual.destroy();

		if (pageMode) {
			document.removeEventListener("scroll", onScroll);
		}
	});

	function getUniqueIdFromDataSources() {
		return data.map(dataSource => dataSource[key]);
	}

	function onItemResized(event) {
		const { id, size, type } = event.detail;

		if (type === "item") virtual.saveSize(id, size); else if (type === "slot") {
			if (id === "header") virtual.updateParam("slotHeaderSize", size); else if (id === "footer") virtual.updateParam("slotFooterSize", size);
		} // virtual.handleSlotSizeChange()
	}

	function onRangeChanged(range_) {
		range = range_;

		$$invalidate(3, paddingStyle = $$invalidate(3, paddingStyle = isHorizontal
		? `0px ${range.padBehind}px 0px ${range.padFront}px`
		: `${range.padFront}px 0px ${range.padBehind}px`));

		$$invalidate(2, displayItems = data.slice(range.start, range.end + 1));
	}

	function onScroll(event) {
		const offset = getOffset();
		const clientSize = getClientSize();
		const scrollSize = getScrollSize();

		// iOS scroll-spring-back behavior will make direction mistake
		if (offset < 0 || offset + clientSize > scrollSize || !scrollSize) {
			return;
		}

		virtual.handleScroll(offset);
		emitEvent(offset, clientSize, scrollSize, event);
	}

	function emitEvent(offset, clientSize, scrollSize, event) {
		dispatch("scroll", { event, range: virtual.getRange() });

		if (virtual.isFront() && !!data.length && offset - topThreshold <= 0) {
			dispatch("top");
		} else if (virtual.isBehind() && offset + clientSize + bottomThreshold >= scrollSize) {
			dispatch("bottom");
		}
	}

	function handleKeepsChange(keeps) {
		virtual.updateParam("keeps", keeps);
		virtual.handleSlotSizeChange();
	}

	async function handleDataSourcesChange(data) {
		virtual.updateParam("uniqueIds", getUniqueIdFromDataSources());
		virtual.handleDataSourcesChange();
	}

	function div1_binding($$value) {
		binding_callbacks[$$value ? "unshift" : "push"](() => {
			shepherd = $$value;
			$$invalidate(5, shepherd);
		});
	}

	function div2_binding($$value) {
		binding_callbacks[$$value ? "unshift" : "push"](() => {
			root = $$value;
			$$invalidate(4, root);
		});
	}

	$$self.$$set = $$props => {
		if ("key" in $$props) $$invalidate(0, key = $$props.key);
		if ("data" in $$props) $$invalidate(9, data = $$props.data);
		if ("keeps" in $$props) $$invalidate(10, keeps = $$props.keeps);
		if ("estimateSize" in $$props) $$invalidate(11, estimateSize = $$props.estimateSize);
		if ("isHorizontal" in $$props) $$invalidate(1, isHorizontal = $$props.isHorizontal);
		if ("start" in $$props) $$invalidate(12, start = $$props.start);
		if ("offset" in $$props) $$invalidate(13, offset = $$props.offset);
		if ("pageMode" in $$props) $$invalidate(14, pageMode = $$props.pageMode);
		if ("topThreshold" in $$props) $$invalidate(15, topThreshold = $$props.topThreshold);
		if ("bottomThreshold" in $$props) $$invalidate(16, bottomThreshold = $$props.bottomThreshold);
		if ("$$scope" in $$props) $$invalidate(29, $$scope = $$props.$$scope);
	};

	$$self.$$.update = () => {
		if ($$self.$$.dirty[0] & /*offset*/ 8192) {
			scrollToOffset(offset);
		}

		if ($$self.$$.dirty[0] & /*start*/ 4096) {
			scrollToIndex(start);
		}

		if ($$self.$$.dirty[0] & /*keeps*/ 1024) {
			handleKeepsChange(keeps);
		}

		if ($$self.$$.dirty[0] & /*data*/ 512) {
			handleDataSourcesChange();
		}
	};

	return [
		key,
		isHorizontal,
		displayItems,
		paddingStyle,
		root,
		shepherd,
		onItemResized,
		onScroll,
		$$slots,
		data,
		keeps,
		estimateSize,
		start,
		offset,
		pageMode,
		topThreshold,
		bottomThreshold,
		getSize,
		getSizes,
		getOffset,
		getClientSize,
		getScrollSize,
		updatePageModeFront,
		scrollToOffset,
		scrollToIndex,
		scrollToBottom,
		slots,
		div1_binding,
		div2_binding,
		$$scope
	];
}

class VirtualScroll extends SvelteComponent {
	constructor(options) {
		super();

		init(
			this,
			options,
			instance,
			create_fragment,
			safe_not_equal,
			{
				key: 0,
				data: 9,
				keeps: 10,
				estimateSize: 11,
				isHorizontal: 1,
				start: 12,
				offset: 13,
				pageMode: 14,
				topThreshold: 15,
				bottomThreshold: 16,
				getSize: 17,
				getSizes: 18,
				getOffset: 19,
				getClientSize: 20,
				getScrollSize: 21,
				updatePageModeFront: 22,
				scrollToOffset: 23,
				scrollToIndex: 24,
				scrollToBottom: 25
			},
			[-1, -1]
		);
	}

	get getSize() {
		return this.$$.ctx[17];
	}

	get getSizes() {
		return this.$$.ctx[18];
	}

	get getOffset() {
		return this.$$.ctx[19];
	}

	get getClientSize() {
		return this.$$.ctx[20];
	}

	get getScrollSize() {
		return this.$$.ctx[21];
	}

	get updatePageModeFront() {
		return this.$$.ctx[22];
	}

	get scrollToOffset() {
		return this.$$.ctx[23];
	}

	get scrollToIndex() {
		return this.$$.ctx[24];
	}

	get scrollToBottom() {
		return this.$$.ctx[25];
	}
}

export default VirtualScroll;
export { VirtualScroll };
//# sourceMappingURL=index.mjs.map
